;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "shell-maker" "0.93.5"
  "Interaction mode for making comint shells."
  '((emacs "27.1"))
  :url "https://github.com/xenodium/shell-maker"
  :commit "071c6df3ca22a2f4c0daa689848ac9bd21bf4e2b"
  :revdesc "071c6df3ca22")
